# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import io
import os
import re
import shutil
import subprocess
import tempfile

import fixtures
import subunit as subunit_lib
import testtools

from stestr.commands import list as list_cmd
from stestr.commands import run
from stestr.tests import base


class TestReturnCodes(base.TestCase):
    def setUp(self):
        super().setUp()
        # Setup test dirs
        self.directory = tempfile.mkdtemp(prefix="stestr-unit")
        self.addCleanup(shutil.rmtree, self.directory, ignore_errors=True)
        self.test_dir = os.path.join(self.directory, "tests")
        os.mkdir(self.test_dir)
        # Setup Test files
        self.repo_root = os.path.abspath(os.curdir)
        self.testr_conf_file = os.path.join(self.directory, ".stestr.conf")
        self.setup_cfg_file = os.path.join(self.directory, "setup.cfg")
        self.passing_file = os.path.join(self.test_dir, "test_passing.py")
        self.failing_file = os.path.join(self.test_dir, "test_failing.py")
        self.init_file = os.path.join(self.test_dir, "__init__.py")
        self.setup_py = os.path.join(self.directory, "setup.py")
        self.user_config = os.path.join(self.directory, "stestr.yaml")
        shutil.copy("stestr/tests/files/testr-conf", self.testr_conf_file)
        shutil.copy("stestr/tests/files/passing-tests", self.passing_file)
        shutil.copy("stestr/tests/files/failing-tests", self.failing_file)
        shutil.copy("stestr/tests/files/setup.py", self.setup_py)
        shutil.copy("stestr/tests/files/setup.cfg", self.setup_cfg_file)
        shutil.copy("stestr/tests/files/__init__.py", self.init_file)
        shutil.copy("stestr/tests/files/stestr.yaml", self.user_config)

        self.stdout = io.StringIO()
        self.stderr = io.StringIO()
        # Change directory, run wrapper and check result
        self.addCleanup(os.chdir, self.repo_root)
        os.chdir(self.directory)
        subprocess.call("stestr init", shell=True)

    def _check_subunit(self, output_stream):
        stream = subunit_lib.ByteStreamToStreamResult(output_stream)
        starts = testtools.StreamResult()
        summary = testtools.StreamSummary()
        tests = []

        def _add_dict(test):
            tests.append(test)

        outcomes = testtools.StreamToDict(functools.partial(_add_dict))
        result = testtools.CopyStreamResult([starts, outcomes, summary])
        result.startTestRun()
        try:
            stream.run(result)
        finally:
            result.stopTestRun()
        self.assertThat(len(tests), testtools.matchers.GreaterThan(0))

    def assertRunExit(self, cmd, expected, subunit=False, stdin=None):
        if stdin:
            p = subprocess.Popen(
                "%s" % cmd,
                shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            out, err = p.communicate(stdin)
        else:
            p = subprocess.Popen(
                "%s" % cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            out, err = p.communicate()

        if not subunit:
            self.assertEqual(
                p.returncode,
                expected,
                "Command: {}; Stdout: {}; Stderr: {}".format(cmd, out, err),
            )
            return (out, err)
        else:
            self.assertEqual(
                p.returncode,
                expected,
                "Expected return code: {} doesn't match actual "
                "return code of: {}. Command: {}".format(expected, p.returncode, cmd),
            )
            output_stream = io.BytesIO(out)
            stream = subunit_lib.ByteStreamToStreamResult(output_stream)
            starts = testtools.StreamResult()
            summary = testtools.StreamSummary()
            tests = []

            def _add_dict(test):
                tests.append(test)

            outcomes = testtools.StreamToDict(functools.partial(_add_dict))
            result = testtools.CopyStreamResult([starts, outcomes, summary])
            result.startTestRun()
            try:
                stream.run(result)
            finally:
                result.stopTestRun()
            self.assertThat(len(tests), testtools.matchers.GreaterThan(0))
            return (out, err)

    def test_parallel_passing(self):
        self.assertRunExit("stestr run passing", 0)

    def test_parallel_passing_bad_regex(self):
        self.assertRunExit("stestr run bad.regex.foobar", 1)

    def test_parallel_fails(self):
        self.assertRunExit("stestr run", 1)

    def test_parallel_passing_xfail(self):
        self.assertRunExit("stestr run xfail", 0)

    def test_parallel_fails_unxsuccess(self):
        self.assertRunExit("stestr run unexpected", 1)

    def test_parallel_exclusion_list(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, "w") as exclusion_list:
            exclusion_list.write("fail")
        cmd = "stestr run --exclude-list %s" % path
        self.assertRunExit(cmd, 0)

    def test_parallel_inclusion_list(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, "w") as inclusion_list:
            inclusion_list.write("passing")
        cmd = "stestr run --include-list %s" % path
        self.assertRunExit(cmd, 0)

    def test_serial_passing(self):
        self.assertRunExit("stestr run --serial passing", 0)

    def test_serial_fails(self):
        self.assertRunExit("stestr run --serial", 1)

    def test_serial_exclusion_list(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, "w") as exclusion_list:
            exclusion_list.write("fail")
        cmd = "stestr run --serial --exclude-list %s" % path
        self.assertRunExit(cmd, 0)

    def test_serial_inclusion_list(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, "w") as inclusion_list:
            inclusion_list.write("passing")
        cmd = "stestr run --serial --include-list %s" % path
        self.assertRunExit(cmd, 0)

    def test_serial_subunit_passing(self):
        self.assertRunExit(
            "stestr --user-config stestr.yaml run --subunit " "--serial passing",
            0,
            subunit=True,
        )

    def test_serial_subunit_failing(self):
        self.assertRunExit(
            "stestr --user-config stestr.yaml run --subunit " "--serial failing",
            0,
            subunit=True,
        )

    def test_parallel_subunit_passing(self):
        self.assertRunExit(
            "stestr --user-config stestr.yaml run --subunit " "passing", 0, subunit=True
        )

    def test_parallel_subunit_failing(self):
        self.assertRunExit(
            "stestr --user-config stestr.yaml run --subunit " "failing", 0, subunit=True
        )

    def test_slowest_passing(self):
        self.assertRunExit("stestr run --slowest passing", 0)

    def test_slowest_failing(self):
        self.assertRunExit("stestr run --slowest failing", 1)

    def test_until_failure_fails(self):
        self.assertRunExit("stestr run --until-failure", 1)

    def test_until_failure_with_subunit_fails(self):
        self.assertRunExit(
            "stestr --user-config stestr.yaml run " "--until-failure --subunit",
            1,
            subunit=True,
        )

    def test_with_parallel_class(self):
        # NOTE(masayukig): Ideally, it's better to figure out the
        # difference between with --parallel-class and without
        # --parallel-class. However, it's difficult to make such a
        # test from a command line based test.
        self.assertRunExit("stestr --parallel-class run passing", 0)

    def test_no_repo_dir(self):
        stestr_repo_dir = os.path.join(self.directory, ".stestr")
        shutil.rmtree(stestr_repo_dir, ignore_errors=True)
        # We can use stestr run even if there's no repo directory.
        self.assertRunExit("stestr run passing", 0)

    def test_empty_repo_dir(self):
        stestr_repo_dir = os.path.join(self.directory, ".stestr")
        shutil.rmtree(stestr_repo_dir, ignore_errors=True)
        os.mkdir(stestr_repo_dir)
        # We can initialize an empty repo directory.
        self.assertRunExit("stestr run passing", 0)

    def test_non_empty_repo_dir(self):
        stestr_repo_dir = os.path.join(self.directory, ".stestr")
        shutil.rmtree(stestr_repo_dir, ignore_errors=True)
        os.mkdir(stestr_repo_dir)
        with open(os.path.join(stestr_repo_dir, "foo"), "wt") as stream:
            stream.write("1\n")
        # We can't initialize a non-empty repo directory.
        self.assertRunExit("stestr run passing", 1)

    def test_list(self):
        self.assertRunExit("stestr list", 0)

    def _get_cmd_stdout(self, cmd):
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
        out = p.communicate()
        self.assertEqual(0, p.returncode)
        return out

    def test_combine_results(self):
        self.assertRunExit("stestr run passing", 0)
        stdout = self._get_cmd_stdout("stestr last --no-subunit-trace")
        stdout = str(stdout[0])
        test_count_split = stdout.split(" ")
        test_count = test_count_split[1]
        test_count = int(test_count)
        id_regex = re.compile(r"\(id=(.*?)\)")
        test_id = id_regex.search(stdout).group(0)
        self.assertRunExit("stestr run --combine passing", 0)
        combine_stdout = self._get_cmd_stdout("stestr last --no-subunit-trace")[0]
        combine_stdout = str(combine_stdout)
        combine_test_count_split = combine_stdout.split(" ")
        combine_test_count = combine_test_count_split[1]
        combine_test_count = int(combine_test_count)
        combine_test_id = id_regex.search(combine_stdout).group(0)
        self.assertEqual(test_id, combine_test_id)

        # The test results from running the same tests twice with combine
        # should return a test count 2x as big at the end of the run
        self.assertEqual(test_count * 2, combine_test_count)

    def test_load_from_stdin(self):
        self.assertRunExit("stestr run passing", 0)
        stream = self._get_cmd_stdout("stestr last --subunit")[0]
        self.assertRunExit("stestr load", 0, stdin=stream)

    def test_load_force_init(self):
        self.assertRunExit("stestr run passing", 0)
        stream = self._get_cmd_stdout("stestr last --subunit")[0]
        # NOTE: --force-init should work here because there is an properly
        # initialized repository.
        self.assertRunExit("stestr load --force-init", 0, stdin=stream)

    def test_load_force_init_invalid(self):
        self.assertRunExit("stestr run passing", 0)
        stream = self._get_cmd_stdout("stestr last --subunit")[0]
        os.remove(os.path.join(self.directory, ".stestr", "format"))
        # NOTE: --force-init should fail here because there is an invalid
        # repository.
        self.assertRunExit("stestr load --force-init", 1, stdin=stream)

    def test_load_from_stdin_quiet(self):
        out, err = self.assertRunExit(
            "stestr --user-config stestr.yaml -q " "run passing", 0
        )
        self.assertEqual(out.decode("utf-8"), "")
        # FIXME(masayukig): We get some warnings when we run a coverage job.
        # So, just ignore 'err' here.
        stream = self._get_cmd_stdout("stestr last --subunit")[0]
        out, err = self.assertRunExit(
            "stestr --user-config stestr.yaml -q " "load", 0, stdin=stream
        )
        self.assertEqual(out.decode("utf-8"), "")
        self.assertEqual(err.decode("utf-8"), "")

    def test_history_list(self):
        self.assertRunExit("stestr run passing", 0)
        self.assertRunExit("stestr run", 1)
        self.assertRunExit("stestr run passing", 0)
        table = self.assertRunExit("stestr history list", 0)[0].decode("utf8")
        self.assertIn("| 0      | True   |", table.split("\n")[3].rstrip())
        self.assertIn("| 1      | False  |", table.split("\n")[4].rstrip())
        self.assertIn("| 2      | True   |", table.split("\n")[5].rstrip())
        expected = """
+--------+--------+-----------+----------------------------------+
| Run ID | Passed | Runtime   | Date                             |
+--------+--------+-----------+----------------------------------+
""".rstrip()
        self.assertEqual(
            expected.strip(),
            "\n".join([x.rstrip() for x in table.split("\n")[:3]]).strip(),
        )

    def test_history_empty(self):
        table = self.assertRunExit("stestr history list", 0)[0].decode("utf8")
        self.assertEqual("", "\n".join([x.rstrip() for x in table.split("\n")]).strip())

    def test_history_show_passing(self):
        self.assertRunExit("stestr run passing", 0)
        self.assertRunExit("stestr run", 1)
        self.assertRunExit("stestr run passing", 0)
        output, _ = self.assertRunExit("stestr history show 0", 0)
        lines = [x.rstrip() for x in output.decode("utf8").split("\n")]
        self.assertIn(" - Passed: 2", lines)
        self.assertIn(" - Failed: 0", lines)
        self.assertIn(" - Expected Fail: 1", lines)

    def test_history_show_failing(self):
        self.assertRunExit("stestr run passing", 0)
        self.assertRunExit("stestr run", 1)
        self.assertRunExit("stestr run passing", 0)
        output, _ = self.assertRunExit("stestr history show 1", 1)
        lines = [x.rstrip() for x in output.decode("utf8").split("\n")]
        self.assertIn(" - Passed: 2", lines)
        self.assertIn(" - Failed: 2", lines)
        self.assertIn(" - Expected Fail: 1", lines)
        self.assertIn(" - Unexpected Success: 1", lines)

    def test_history_show_invalid_id(self):
        self.assertRunExit("stestr run passing", 0)
        self.assertRunExit("stestr run", 1)
        self.assertRunExit("stestr run passing", 0)
        output, _ = self.assertRunExit("stestr history show 42", 1)
        self.assertEqual(output.decode("utf8").rstrip(), "'No such run.'")

    def test_history_remove(self):
        self.assertRunExit("stestr run passing", 0)
        self.assertRunExit("stestr run", 1)
        self.assertRunExit("stestr run passing", 0)
        self.assertRunExit("stestr history remove 1", 0)
        table = self.assertRunExit("stestr history list", 0)[0].decode("utf8")
        self.assertIn("| 0      | True   |", table.split("\n")[3].rstrip())
        self.assertNotIn("| 1      | False  |", table.split("\n")[4].strip())
        self.assertIn("| 2      | True   |", table.split("\n")[4].rstrip())
        expected = """
+--------+--------+-----------+----------------------------------+
| Run ID | Passed | Runtime   | Date                             |
+--------+--------+-----------+----------------------------------+
""".strip()
        self.assertEqual(
            expected, "\n".join([x.rstrip() for x in table.split("\n")[:3]])
        )

    def test_no_subunit_trace_force_subunit_trace(self):
        out, err = self.assertRunExit(
            "stestr run --no-subunit-trace --force-subunit-trace passing", 0
        )
        out = str(out)
        self.assertNotIn("PASSED (id=0)", out)
        self.assertIn("Totals", out)
        self.assertIn("Worker Balance", out)
        self.assertIn("Sum of execute time for each test:", out)

    def test_parallel_passing_from_func(self):
        stdout = fixtures.StringStream("stdout")
        self.useFixture(stdout)
        self.assertEqual(0, run.run_command(filters=["passing"], stdout=stdout.stream))

    def test_parallel_passing_bad_regex_from_func(self):
        stdout = fixtures.StringStream("stdout")
        self.useFixture(stdout)
        self.assertEqual(
            1, run.run_command(filters=["bad.regex.foobar"], stdout=stdout.stream)
        )

    def test_parallel_fails_from_func(self):
        stdout = fixtures.StringStream("stdout")
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(stdout=stdout.stream))

    def test_serial_passing_from_func(self):
        stdout = fixtures.StringStream("stdout")
        self.useFixture(stdout)
        self.assertEqual(
            0, run.run_command(filters=["passing"], serial=True, stdout=stdout.stream)
        )

    def test_str_concurrency_passing_from_func(self):
        stdout = fixtures.StringStream("stdout")
        self.useFixture(stdout)
        self.assertEqual(
            0,
            run.run_command(filters=["passing"], concurrency="1", stdout=stdout.stream),
        )

    def test_str_concurrency_fails_from_func(self):
        stdout = fixtures.StringStream("stdout")
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(concurrency="1", stdout=stdout.stream))

    def test_serial_fails_from_func(self):
        stdout = fixtures.StringStream("stdout")
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(serial=True, stdout=stdout.stream))

    def test_serial_subunit_passing_from_func(self):
        stdout = io.BytesIO()
        self.assertEqual(
            0,
            run.run_command(
                subunit_out=True, serial=True, filters=["passing"], stdout=stdout
            ),
        )
        stdout.seek(0)
        self._check_subunit(stdout)

    def test_parallel_subunit_passing_from_func(self):
        stdout = io.BytesIO()
        self.assertEqual(
            0, run.run_command(subunit_out=True, filters=["passing"], stdout=stdout)
        )
        stdout.seek(0)
        self._check_subunit(stdout)

    def test_until_failure_fails_from_func(self):
        stdout = fixtures.StringStream("stdout")
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(until_failure=True, stdout=stdout.stream))

    def test_until_failure_with_subunit_fails_from_func(self):
        stdout = io.BytesIO()
        self.assertEqual(
            1, run.run_command(until_failure=True, subunit_out=True, stdout=stdout)
        )
        stdout.seek(0)
        self._check_subunit(stdout)

    def test_list_from_func(self):
        stdout = fixtures.StringStream("stdout")
        self.useFixture(stdout)
        self.assertEqual(0, list_cmd.list_command(stdout=stdout.stream))

    def test_run_no_discover_pytest_path(self):
        passing_string = "::".join(
            [
                os.path.join("tests", "test_passing.py"),
                "FakeTestClass",
                "test_pass_list",
            ]
        )
        out, err = self.assertRunExit("stestr run -n %s" % passing_string, 0)
        lines = out.decode("utf8").splitlines()
        self.assertIn(" - Passed: 1", lines)
        self.assertIn(" - Failed: 0", lines)

    def test_run_no_discover_pytest_path_failing(self):
        failing_string = "::".join(
            [
                os.path.join("tests", "test_failing.py"),
                "FakeTestClass",
                "test_pass_list",
            ]
        )
        out, err = self.assertRunExit("stestr run -n %s" % failing_string, 1)
        lines = out.decode("utf8").splitlines()
        self.assertIn(" - Passed: 0", lines)
        self.assertIn(" - Failed: 1", lines)

    def test_run_no_discover_file_path(self):
        passing_string = os.path.join("tests", "test_passing.py")
        out, err = self.assertRunExit("stestr run -n %s" % passing_string, 0)
        lines = out.decode("utf8").splitlines()
        self.assertIn(" - Passed: 2", lines)
        self.assertIn(" - Failed: 0", lines)
        self.assertIn(" - Expected Fail: 1", lines)

    def test_run_no_discover_file_path_failing(self):
        failing_string = os.path.join("tests", "test_failing.py")
        out, err = self.assertRunExit("stestr run -n %s" % failing_string, 1)
        lines = out.decode("utf8").splitlines()
        self.assertIn(" - Passed: 0", lines)
        self.assertIn(" - Failed: 2", lines)
        self.assertIn(" - Unexpected Success: 1", lines)

    def test_run_no_discover_unnormalized_file_path(self):
        # we don't use os.path.join since we want an unnormalized path
        passing_string = f"tests{os.path.sep}{os.path.sep}test_passing.py"
        out, err = self.assertRunExit("stestr run -n %s" % passing_string, 0)
        lines = out.decode("utf8").splitlines()
        self.assertIn(" - Passed: 2", lines)
        self.assertIn(" - Failed: 0", lines)
        self.assertIn(" - Expected Fail: 1", lines)

    def test_run_no_discover_unnormalized_file_path_failing(self):
        # we don't use os.path.join since we want an unnormalized path
        failing_string = f"tests{os.path.sep}{os.path.sep}test_failing.py"
        out, err = self.assertRunExit("stestr run -n %s" % failing_string, 1)
        lines = out.decode("utf8").splitlines()
        self.assertIn(" - Passed: 0", lines)
        self.assertIn(" - Failed: 2", lines)
        self.assertIn(" - Unexpected Success: 1", lines)


class TestReturnCodesToxIni(TestReturnCodes):
    def setUp(self):
        super().setUp()
        self.tox_ini_dir = os.path.join(self.directory, "tox.ini")
        tox_file = os.path.join(self.repo_root, "stestr/tests/files/tox.ini")
        shutil.copy(tox_file, self.tox_ini_dir)
        os.remove(self.testr_conf_file)


class TestReturnCodesTOML(TestReturnCodes):
    def setUp(self):
        super().setUp()
        self.pyproject_toml = os.path.join(self.directory, "pyproject.toml")
        toml_conf = os.path.join(self.repo_root, "stestr/tests/files/conf.toml")
        shutil.copy(toml_conf, self.pyproject_toml)
        os.remove(self.testr_conf_file)

    def test_all_configs_missing(self):
        stestr_repo_dir = os.path.join(self.directory, ".stestr")
        shutil.rmtree(stestr_repo_dir, ignore_errors=True)
        os.remove(self.pyproject_toml)
        output, _ = self.assertRunExit("stestr run passing", 1)
        self.assertIn(b"No config file found", output)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import itertools
import multiprocessing
import operator
import random

import yaml

from stestr import selection


def partition_tests(test_ids, concurrency, repository, group_callback, randomize=False):
    """Partition test_ids by concurrency.

    Test durations from the repository are used to get partitions which
    have roughly the same expected runtime. New tests - those with no
    recorded duration - are allocated in round-robin fashion to the
    partitions created using test durations.

    :param list test_ids: The list of test_ids to be partitioned
    :param int concurrency: The concurrency that will be used for running
        the tests. This is the number of partitions that test_ids will be
        split into.
    :param repository: A repository object that
    :param group_callback: A callback function that is used as a scheduler
        hint to group test_ids together and treat them as a single unit for
        scheduling. This function expects a single test_id parameter and it
        will return a group identifier. Tests_ids that have the same group
        identifier will be kept on the same worker.
    :param bool randomize: If true each partition's test order will be
                        randomized

    :return: A list where each element is a distinct subset of test_ids,
        and the union of all the elements is equal to set(test_ids).
    """

    def noop(_):
        return None

    _group_callback = group_callback
    partitions = [list() for i in range(concurrency)]
    timed_partitions = [[0.0, partition] for partition in partitions]
    time_data = {}
    if repository:
        time_data = repository.get_test_times(test_ids)
        timed_tests = time_data["known"]
        unknown_tests = time_data["unknown"]
    else:
        timed_tests = {}
        unknown_tests = set(test_ids)
    # Group tests: generate group_id -> test_ids.
    group_ids = collections.defaultdict(list)
    if _group_callback is None:
        group_callback = noop
    else:
        group_callback = _group_callback
    for test_id in test_ids:
        group_id = group_callback(test_id) or test_id
        group_ids[group_id].append(test_id)
    # Time groups: generate three sets of groups:
    # - fully timed dict(group_id -> time),
    # - partially timed dict(group_id -> time) and
    # - unknown (set of group_id)
    # We may in future treat partially timed different for scheduling, but
    # at least today we just schedule them after the fully timed groups.
    timed = {}
    partial = {}
    unknown = []
    for group_id, group_tests in group_ids.items():
        untimed_ids = unknown_tests.intersection(group_tests)
        group_time = sum(
            [
                timed_tests[test_id]
                for test_id in untimed_ids.symmetric_difference(group_tests)
            ]
        )
        if not untimed_ids:
            timed[group_id] = group_time
        elif group_time:
            partial[group_id] = group_time
        else:
            unknown.append(group_id)

    # Scheduling is NP complete in general, so we avoid aiming for
    # perfection. A quick approximation that is sufficient for our general
    # needs:
    # sort the groups by time
    # allocate to partitions by putting each group in to the partition with
    # the current (lowest time, shortest length[in tests])
    def consume_queue(groups):
        queue = sorted(groups.items(), key=operator.itemgetter(1), reverse=True)
        for group_id, duration in queue:
            timed_partitions[0][0] = timed_partitions[0][0] + duration
            timed_partitions[0][1].extend(group_ids[group_id])
            timed_partitions.sort(key=lambda item: (item[0], len(item[1])))

    consume_queue(timed)
    consume_queue(partial)
    # Assign groups with entirely unknown times in round robin fashion to
    # the partitions.
    for partition, group_id in zip(itertools.cycle(partitions), unknown):
        partition.extend(group_ids[group_id])
    if randomize:
        out_parts = []
        for partition in partitions:
            temp_part = list(partition)
            random.shuffle(temp_part)
            out_parts.append(list(temp_part))
        return out_parts
    else:
        return partitions


def local_concurrency():
    """Get the number of available CPUs on the system.

    :return: An int for the number of cpus. Or None if it couldn't be found
    """

    try:
        return multiprocessing.cpu_count()
    except NotImplementedError:
        # No concurrency logic known.
        return None


def generate_worker_partitions(
    ids, worker_path, repository=None, group_callback=None, randomize=False
):
    """Parse a worker yaml file and generate test groups

    :param list ids: A list of test ids too be partitioned
    :param path worker_path: The path to a worker file
    :param repository: A repository object that will be used for looking up
        timing data. This is optional, and also will only be used for
        scheduling if there is a count field on a worker.
    :param group_callback: A callback function that is used as a scheduler
        hint to group test_ids together and treat them as a single unit for
        scheduling. This function expects a single test_id parameter and it
        will return a group identifier. Tests_ids that have the same group
        identifier will be kept on the same worker. This is optional and
        also will only be used for scheduling if there is a count field on a
        worker.
    :param bool randomize: If true each partition's test order will be
        randomized. This is optional and also will only be used for scheduling
        if there is a count field on a worker.

    :returns: A list where each element is a distinct subset of test_ids.
    """
    with open(worker_path) as worker_file:
        workers_desc = yaml.safe_load(worker_file.read())
    worker_groups = []
    for worker in workers_desc:
        if isinstance(worker, dict) and "worker" in worker.keys():
            if isinstance(worker["worker"], list):
                local_worker_list = selection.filter_tests(worker["worker"], ids)

                if "concurrency" in worker.keys() and worker["concurrency"] > 1:
                    partitioned_tests = partition_tests(
                        local_worker_list,
                        worker["concurrency"],
                        repository,
                        group_callback,
                        randomize,
                    )
                    worker_groups.extend(partitioned_tests)
                else:
                    # If a worker partition is empty don't add it to the output
                    if local_worker_list:
                        worker_groups.append(local_worker_list)
            else:
                raise TypeError("The input yaml is the incorrect format")
        else:
            raise TypeError("The input yaml is the incorrect format")
    return worker_groups



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import operator
from unittest import mock

import subunit
import testtools

from stestr import bisect_tests
from stestr.repository import abstract
from stestr.tests import base


class FakeTestRun(abstract.AbstractTestRun):
    def get_test(self):
        case = subunit.ByteStreamToStreamResult(io.BytesIO(self._content))

        def wrap_result(result):
            # Wrap in a router to mask out startTestRun/stopTestRun from the
            # ExtendedToStreamDecorator.
            result = testtools.StreamResultRouter(result, do_start_stop_run=False)
            # Wrap that in ExtendedToStreamDecorator to convert v1 calls to
            # StreamResult.
            return testtools.ExtendedToStreamDecorator(result)

        return testtools.DecorateTestCaseResult(
            case,
            wrap_result,
            operator.methodcaller("startTestRun"),
            operator.methodcaller("stopTestRun"),
        )

    def get_id(self):
        return self.id


class FakeFailedTestRunNoTags(FakeTestRun):
    def __init__(self, failure=True):
        # Generate a subunit stream
        self.id = 2
        stream_buf = io.BytesIO()
        stream = subunit.StreamResultToBytes(stream_buf)
        stream.status(test_id="test_a", test_status="inprogress")
        stream.status(test_id="test_a", test_status="success")
        stream.status(test_id="test_b", test_status="inprogress")
        stream.status(test_id="test_b", test_status="success")
        stream.status(test_id="test_c", test_status="inprogress")
        stream.status(test_id="test_c", test_status="fail")
        stream_buf.seek(0)
        self._content = stream_buf.getvalue()


class FakeFailingWithTags(FakeTestRun):
    def __init__(self, failure=True):
        # Generate a subunit stream
        self.id = None
        stream_buf = io.BytesIO()
        stream = subunit.StreamResultToBytes(stream_buf)
        stream.status(
            test_id="test_c", test_status="inprogress", test_tags=["worker-0"]
        )
        stream.status(test_id="test_c", test_status="fail", test_tags=["worker-0"])
        stream_buf.seek(0)
        self._content = stream_buf.getvalue()


class FakeNoFailing(FakeTestRun):
    def __init__(self, failure=True):
        # Generate a subunit stream
        stream_buf = io.BytesIO(bytes(b""))
        self._content = stream_buf.getvalue()
        self.id = None


class FakeFailedTestRunWithTags(FakeTestRun):
    def __init__(self, failure=True):
        # Generate a subunit stream
        stream_buf = io.BytesIO()
        stream = subunit.StreamResultToBytes(stream_buf)
        stream.status(
            test_id="test_a", test_status="inprogress", test_tags=["worker-0"]
        )
        stream.status(test_id="test_a", test_status="success", test_tags=["worker-0"])
        stream.status(
            test_id="test_b", test_status="inprogress", test_tags=["worker-1"]
        )
        stream.status(test_id="test_b", test_status="success", test_tags=["worker-1"])
        stream.status(
            test_id="test_c", test_status="inprogress", test_tags=["worker-0"]
        )
        stream.status(test_id="test_c", test_status="fail", test_tags=["worker-0"])
        stream_buf.seek(0)
        self._content = stream_buf.getvalue()
        self.id = 2


class FakeFailedMultiWorkerTestRunWithTags(FakeTestRun):
    def __init__(self, failure=True):
        # Generate a subunit stream
        stream_buf = io.BytesIO()
        stream = subunit.StreamResultToBytes(stream_buf)
        stream.status(
            test_id="test_a", test_status="inprogress", test_tags=["worker-0"]
        )
        stream.status(test_id="test_a", test_status="success", test_tags=["worker-0"])
        stream.status(
            test_id="test_b", test_status="inprogress", test_tags=["worker-1"]
        )
        stream.status(test_id="test_b", test_status="fail", test_tags=["worker-1"])
        stream.status(
            test_id="test_c", test_status="inprogress", test_tags=["worker-0"]
        )
        stream.status(test_id="test_c", test_status="fail", test_tags=["worker-0"])
        stream_buf.seek(0)
        self._content = stream_buf.getvalue()
        self.id = 2


class TestBisectTests(base.TestCase):
    def setUp(self):
        super().setUp()
        self.repo_mock = mock.create_autospec("stestr.repository.file.Repository")
        self.conf_mock = mock.create_autospec("stestr.config_file.TestrConf")
        self.run_func_mock = mock.MagicMock()
        self.latest_run_mock = mock.MagicMock()

    def test_bisect_no_failures_provided(self):
        bisector = bisect_tests.IsolationAnalyzer(
            self.latest_run_mock, self.conf_mock, self.run_func_mock, self.repo_mock
        )
        self.assertRaises(ValueError, bisector.bisect_tests, [])

    def test_prior_tests_invlaid_test_id(self):
        bisector = bisect_tests.IsolationAnalyzer(
            self.latest_run_mock, self.conf_mock, self.run_func_mock, self.repo_mock
        )
        run = FakeFailedTestRunNoTags()
        self.assertRaises(KeyError, bisector._prior_tests, run, "bad_test_id")

    def test_get_prior_tests_no_tags(self):
        bisector = bisect_tests.IsolationAnalyzer(
            self.latest_run_mock, self.conf_mock, self.run_func_mock, self.repo_mock
        )
        run = FakeFailedTestRunNoTags()
        prior_tests = bisector._prior_tests(run, "test_c")
        self.assertEqual(["test_a", "test_b"], prior_tests)

    def test_get_prior_tests_with_tags(self):
        bisector = bisect_tests.IsolationAnalyzer(
            self.latest_run_mock, self.conf_mock, self.run_func_mock, self.repo_mock
        )
        run = FakeFailedTestRunWithTags()
        prior_tests = bisector._prior_tests(run, "test_c")
        self.assertEqual(["test_a"], prior_tests)

    @mock.patch("stestr.output.output_table")
    def test_bisect_tests_isolated_failure(self, table_mock):
        run = FakeFailedTestRunWithTags()
        self.conf_mock.get_run_command = mock.MagicMock()

        def get_failures(*args, **kwargs):
            return FakeNoFailing()

        self.repo_mock.get_failing = get_failures
        bisector = bisect_tests.IsolationAnalyzer(
            run, self.conf_mock, self.run_func_mock, self.repo_mock
        )
        return_code = bisector.bisect_tests(["test_c"])
        expected_issue = [
            ("failing test", "caused by test"),
            ("test_c", "unknown - no conflicts"),
        ]
        table_mock.assert_called_once_with(expected_issue)
        self.assertEqual(3, return_code)

    @mock.patch("stestr.output.output_table")
    def test_bisect_tests_not_isolated_failure(self, table_mock):
        run = FakeFailedTestRunWithTags()
        self.conf_mock.get_run_command = mock.MagicMock()

        def get_failures(*args, **kwargs):
            return FakeFailingWithTags()

        self.repo_mock.get_failing = get_failures
        bisector = bisect_tests.IsolationAnalyzer(
            run, self.conf_mock, self.run_func_mock, self.repo_mock
        )
        return_code = bisector.bisect_tests(["test_c"])
        expected_issue = [("failing test", "caused by test"), ("test_c", "test_a")]
        table_mock.assert_called_once_with(expected_issue)
        self.assertEqual(3, return_code)

    @mock.patch("stestr.output.output_table")
    def test_bisect_tests_not_isolated_multiworker_failures(self, table_mock):
        run = FakeFailedMultiWorkerTestRunWithTags()
        self.conf_mock.get_run_command = mock.MagicMock()

        def get_failures(*args, **kwargs):
            return FakeFailingWithTags()

        self.repo_mock.get_failing = get_failures
        bisector = bisect_tests.IsolationAnalyzer(
            run, self.conf_mock, self.run_func_mock, self.repo_mock
        )
        return_code = bisector.bisect_tests(["test_b", "test_c"])
        expected_issue = [
            ("failing test", "caused by test"),
            ("test_b", "unknown - no conflicts"),
            ("test_c", "test_a"),
        ]
        table_mock.assert_called_once_with(expected_issue)
        self.assertEqual(3, return_code)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Show the last run loaded into a repository."""

import sys

from cliff import command

from stestr import output
from stestr.repository import abstract
from stestr.repository import util
from stestr import results
from stestr import subunit_trace
from stestr import user_config


class Last(command.Command):
    """Show the last run loaded into a repository.

    Failing tests are shown on the console and a summary of the run is
    printed at the end.

    Without --subunit, the process exit code will be non-zero if the test
    run was not successful. With --subunit, the process exit code is
    non-zero only if the subunit stream could not be generated
    successfully. The test results and run status are included in the
    subunit stream, so the stream should be used to determining the result
    of the run instead of the exit code when using the --subunit flag.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--subunit",
            action="store_true",
            default=False,
            help="Show output as a subunit stream.",
        )
        parser.add_argument(
            "--no-subunit-trace",
            action="store_true",
            default=False,
            help="Disable output with the subunit-trace " "output filter",
        )
        parser.add_argument(
            "--force-subunit-trace",
            action="store_true",
            default=False,
            help="Force subunit-trace output regardless of any"
            "other options or config settings",
        )
        parser.add_argument(
            "--color",
            action="store_true",
            default=False,
            help="Enable color output in the subunit-trace "
            "output, if subunit-trace output is enabled. "
            "(this is the default). If subunit-trace is "
            "disable this does nothing.",
        )
        parser.add_argument(
            "--suppress-attachments",
            action="store_true",
            dest="suppress_attachments",
            help="If set do not print stdout or stderr "
            "attachment contents on a successful test "
            "execution",
        )
        parser.add_argument(
            "--all-attachments",
            action="store_true",
            dest="all_attachments",
            help="If set print all text attachment contents on"
            " a successful test execution",
        )
        parser.add_argument(
            "--show-binary-attachments",
            action="store_true",
            dest="show_binary_attachments",
            help="If set, show non-text attachments. This is "
            "generally only useful for debug purposes.",
        )
        return parser

    def take_action(self, parsed_args):
        user_conf = user_config.get_user_config(self.app_args.user_config)
        args = parsed_args
        if args.suppress_attachments and args.all_attachments:
            msg = (
                "The --suppress-attachments and --all-attachments "
                "options are mutually exclusive, you can not use both "
                "at the same time"
            )
            print(msg)
            sys.exit(1)
        if getattr(user_conf, "last", False):
            if not user_conf.last.get("no-subunit-trace"):
                if not args.no_subunit_trace:
                    pretty_out = True
                else:
                    pretty_out = False
            else:
                pretty_out = False
            pretty_out = args.force_subunit_trace or pretty_out
            color = args.color or user_conf.last.get("color", False)
            suppress_attachments_conf = user_conf.run.get("suppress-attachments", False)
            all_attachments_conf = user_conf.run.get("all-attachments", False)
            if not args.suppress_attachments and not args.all_attachments:
                suppress_attachments = suppress_attachments_conf
                all_attachments = all_attachments_conf
            elif args.suppress_attachments:
                all_attachments = False
                suppress_attachments = args.suppress_attachments
            elif args.all_attachments:
                suppress_attachments = False
                all_attachments = args.all_attachments
        else:
            pretty_out = args.force_subunit_trace or not args.no_subunit_trace
            color = args.color
            suppress_attachments = args.suppress_attachments
            all_attachments = args.all_attachments
        return last(
            repo_url=self.app_args.repo_url,
            subunit_out=args.subunit,
            pretty_out=pretty_out,
            color=color,
            suppress_attachments=suppress_attachments,
            all_attachments=all_attachments,
            show_binary_attachments=args.show_binary_attachments,
        )


def last(
    repo_url=None,
    subunit_out=False,
    pretty_out=True,
    color=False,
    stdout=sys.stdout,
    suppress_attachments=False,
    all_attachments=False,
    show_binary_attachments=False,
):
    """Show the last run loaded into a a repository

    This function will print the results from the last run in the repository
    to STDOUT. It can optionally print the subunit stream for the last run
    to STDOUT if the ``subunit`` option is set to true.

    Note this function depends on the cwd for the repository if `repo_url` is
    not specified it will use the repository located at CWD/.stestr

    :param str repo_url: The url of the repository to use.
    :param bool subunit_out: Show output as a subunit stream.
    :param pretty_out: Use the subunit-trace output filter.
    :param color: Enable colorized output with the subunit-trace output filter.
    :param bool subunit: Show output as a subunit stream.
    :param file stdout: The output file to write all output to. By default
         this is sys.stdout
    :param bool suppress_attachments: When set true attachments subunit_trace
        will not print attachments on successful test execution.
    :param bool all_attachments: When set true subunit_trace will print all
        text attachments on successful test execution.
    :param bool show_binary_attachments: When set to true, subunit_trace will
        print binary attachments in addition to text attachments.

    :return return_code: The exit code for the command. 0 for success and > 0
        for failures.
    :rtype: int
    """
    try:
        repo = util.get_repo_open(repo_url=repo_url)
    except abstract.RepositoryNotFound as e:
        stdout.write(str(e) + "\n")
        return 1

    try:
        latest_run = repo.get_latest_run()
    except KeyError as e:
        stdout.write(str(e) + "\n")
        return 1

    if subunit_out:
        stream = latest_run.get_subunit_stream()
        output.output_stream(stream, output=stdout)
        # Exits 0 if we successfully wrote the stream.
        return 0
    case = latest_run.get_test()
    try:
        previous_run = repo.get_test_run(repo.latest_id() - 1)
    except KeyError:
        previous_run = None
    failed = False
    if not pretty_out:
        output_result = results.CLITestResult(latest_run.get_id, stdout, previous_run)
        summary = output_result.get_summary()
        output_result.startTestRun()
        try:
            case.run(output_result)
        finally:
            output_result.stopTestRun()
        failed = not results.wasSuccessful(summary)
    else:
        stream = latest_run.get_subunit_stream()
        failed = subunit_trace.trace(
            stream,
            stdout,
            post_fails=True,
            color=color,
            suppress_attachments=suppress_attachments,
            all_attachments=all_attachments,
            show_binary_attachments=show_binary_attachments,
        )
    if failed:
        return 1
    else:
        return 0



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import os
from unittest import mock

from stestr.tests import base
from stestr import user_config

FULL_YAML = """
run:
  concurrency: 42 # This can be any integer value
  random: True
  no-subunit-trace: True
  color: True
  abbreviate: True
  slowest: True
  suppress-attachments: True
  all-attachments: True
failing:
  list: True
last:
  no-subunit-trace: True
  color: True
  suppress-attachments: True
  all-attachments: True
load:
  force-init: True
  subunit-trace: True
  color: True
  abbreviate: True
  suppress-attachments: True
  all-attachments: True
history-list:
  show-metadata: True
history-show:
  no-subunit-trace: True
  color: True
  suppress-attachments: True
  all-attachments: True
"""

INVALID_YAML_FIELD = """
run:
  color: True,
"""

YAML_NOT_INT = """
run:
  concurrency: Two
"""


class TestUserConfig(base.TestCase):
    def setUp(self):
        super().setUp()
        home_dir = os.path.expanduser("~")
        self.xdg_path = os.path.join(os.path.join(home_dir, ".config"), "stestr.yaml")
        self.home_path = os.path.join(home_dir, ".stestr.yaml")

    @mock.patch("sys.exit")
    @mock.patch("stestr.user_config.UserConfig")
    def test_get_user_config_invalid_path(self, user_mock, exit_mock):
        user_config.get_user_config("/i_am_an_invalid_path")
        msg = "The specified stestr user config is not a valid path"
        exit_mock.assert_called_once_with(msg)

    @mock.patch("os.path.isfile")
    @mock.patch("stestr.user_config.UserConfig")
    def test_get_user_config_xdg_file(self, user_mock, path_mock):
        def fake_isfile(path):
            if path == self.xdg_path:
                return True
            else:
                return False

        path_mock.side_effect = fake_isfile
        user_config.get_user_config()
        user_mock.assert_called_once_with(self.xdg_path)

    @mock.patch("os.path.isfile")
    @mock.patch("stestr.user_config.UserConfig")
    def test_get_default_user_config_file(self, user_mock, path_mock):
        def fake_isfile(path):
            if path == self.home_path:
                return True
            else:
                return False

        path_mock.side_effect = fake_isfile
        user_config.get_user_config()
        user_mock.assert_called_once_with(self.home_path)

    @mock.patch("yaml.safe_load", return_value={})
    @mock.patch("builtins.open", mock.mock_open())
    def test_user_config_empty_schema(self, yaml_mock):
        user_conf = user_config.UserConfig("/path")
        self.assertEqual({}, user_conf.config)

    @mock.patch("yaml.safe_load", return_value={"init": {"subunit-trace": True}})
    @mock.patch("sys.exit")
    @mock.patch("builtins.open", mock.mock_open())
    def test_user_config_invalid_command(self, exit_mock, yaml_mock):
        user_config.UserConfig("/path")
        error_string = (
            "Provided user config file /path is invalid because:\n"
            "extra keys not allowed @ data['init']"
        )
        exit_mock.assert_called_once_with(error_string)

    @mock.patch("yaml.safe_load", return_value={"run": {"subunit-trace": True}})
    @mock.patch("sys.exit")
    @mock.patch("builtins.open", mock.mock_open())
    def test_user_config_invalid_option(self, exit_mock, yaml_mock):
        user_config.UserConfig("/path")
        error_string = (
            "Provided user config file /path is invalid because:\n"
            "extra keys not allowed @ "
            "data['run']['subunit-trace']"
        )
        exit_mock.assert_called_once_with(error_string)

    @mock.patch("builtins.open", return_value=io.BytesIO(FULL_YAML.encode("utf-8")))
    def test_user_config_full_config(self, open_mock):
        user_conf = user_config.UserConfig("/path")
        full_dict = {
            "run": {
                "concurrency": 42,
                "random": True,
                "no-subunit-trace": True,
                "color": True,
                "abbreviate": True,
                "slowest": True,
                "suppress-attachments": True,
                "all-attachments": True,
            },
            "failing": {"list": True},
            "last": {
                "no-subunit-trace": True,
                "color": True,
                "suppress-attachments": True,
                "all-attachments": True,
            },
            "load": {
                "force-init": True,
                "subunit-trace": True,
                "color": True,
                "abbreviate": True,
                "suppress-attachments": True,
                "all-attachments": True,
            },
            "history-list": {"show-metadata": True},
            "history-show": {
                "no-subunit-trace": True,
                "color": True,
                "suppress-attachments": True,
                "all-attachments": True,
            },
        }
        self.assertEqual(full_dict, user_conf.config)

    @mock.patch("sys.exit")
    @mock.patch(
        "builtins.open", return_value=io.BytesIO(INVALID_YAML_FIELD.encode("utf-8"))
    )
    def test_user_config_invalid_value_type(self, open_mock, exit_mock):
        user_config.UserConfig("/path")
        error_string = (
            "Provided user config file /path is invalid because:\n"
            "expected bool for dictionary value @ "
            "data['run']['color']"
        )
        exit_mock.assert_called_once_with(error_string)

    @mock.patch("sys.exit")
    @mock.patch("builtins.open", return_value=io.BytesIO(YAML_NOT_INT.encode("utf-8")))
    def test_user_config_invalid_integer(self, open_mock, exit_mock):
        user_config.UserConfig("/path")
        error_string = (
            "Provided user config file /path is invalid because:\n"
            "expected int for dictionary value @ "
            "data['run']['concurrency']"
        )
        exit_mock.assert_called_once_with(error_string)
